@@ -5,11 +5,15 @@ from .core import console, device_select, load_images, progress_bar
 from .models import EmbeddingExtractor
 from PIL import Image
 import matplotlib.pyplot as plt
-from rich.table import Table
+from sklearn.decomposition import PCA
+from sklearn.manifold import TSNE
+import matplotlib.pyplot as plt
+
 
 class CleanFrame:
     def __init__(self, path, model="clip", cluster="kmeans", n_clusters=5,
-                 device="auto", batch_size=32, eps=0.5, min_samples=5, threshold=0.95):
+                 device="auto", batch_size=32, eps=0.5, min_samples=5,
+                 threshold=0.95, cache=True):
 
         self.path = path
         self.model = model.lower()
@@ -19,16 +23,22 @@ class CleanFrame:
         self.min_samples = min_samples
         self.batch_size = batch_size
         self.threshold = threshold
+        self.cache = cache
 
         # Device
         self.device = device_select(device)
-        console.print(f"[bold cyan]Device:[/bold cyan] {self.device}")
+        console.print(f"Device: {self.device}")
 
-        # Directories
+        # Paths
         self.dataset_name = os.path.basename(os.path.normpath(path))
-        self.root = os.path.join(".cleanframe_cache", self.dataset_name)
-        self.cleaned_dir = os.path.join(self.root, "cleaned")
+        base_dir = os.path.dirname(os.path.normpath(path))
+        self.cleaned_dir = os.path.join(base_dir, f"{self.dataset_name}_cleaned")
+        self.removed_dir = os.path.join(base_dir, f"{self.dataset_name}_removed")
+        self.emb_cache_path = os.path.join(base_dir, f"{self.dataset_name}_embeddings.npy")
+        self.report_path = os.path.join(base_dir, f"{self.dataset_name}_report.txt")
+
         os.makedirs(self.cleaned_dir, exist_ok=True)
+        os.makedirs(self.removed_dir, exist_ok=True)
 
         # Embedding extractor
         self.extractor = EmbeddingExtractor(self.device, self.batch_size)
@@ -40,12 +50,36 @@ class CleanFrame:
         self.cleaned_paths = []
         self.removed_paths = []
 
+    # ================================================================
+    # Embedding Management (with cache)
+    # ================================================================
+
+    def _get_or_load_embeddings(self):
+        if self.cache and os.path.exists(self.emb_cache_path):
+            console.print(f"Loading cached embeddings from {self.emb_cache_path}")
+            self.embeddings = np.load(self.emb_cache_path)
+            return self.embeddings, self.paths
+
+        console.print("Computing embeddings (no cache found)...")
+        images, paths = load_images(self.path)
+        if not paths:
+            raise FileNotFoundError(f"No images found in {self.path}")
+
+        self.paths = paths
+        self.embeddings = self.extractor.get_embeddings(images, self.model)
+
+        if self.cache:
+            np.save(self.emb_cache_path, self.embeddings)
+            console.print(f"Embeddings cached at: {self.emb_cache_path}")
+
+        return self.embeddings, self.paths
+
     # ================================================================
     # Clustering and Cleaning
     # ================================================================
 
     def _cluster(self, embeddings):
-        console.print(f"[cyan]Clustering using {self.cluster_method.upper()}...[/cyan]")
+        console.print(f"Clustering using {self.cluster_method.upper()}...")
         if self.cluster_method == "kmeans":
             labels = KMeans(n_clusters=self.n_clusters, random_state=42, n_init="auto").fit_predict(embeddings)
         elif self.cluster_method == "dbscan":
@@ -77,6 +111,7 @@ class CleanFrame:
                         kept.append(cluster_paths[i])
                     else:
                         removed.append(cluster_paths[i])
+                        shutil.copy(cluster_paths[i], os.path.join(self.removed_dir, os.path.basename(cluster_paths[i])))
                 p.update(task, advance=1)
 
         for f in kept:
@@ -84,11 +119,12 @@ class CleanFrame:
 
         self.cleaned_paths = kept
         self.removed_paths = removed
-        console.print(f"[bold cyan]Kept:[/bold cyan] {len(kept)}   [bold red]Removed:[/bold red] {len(removed)}")
-        console.print(f"[yellow]Output → {self.cleaned_dir}[/yellow]")
+        console.print(f"Kept: {len(kept)}   Removed: {len(removed)}")
+        console.print(f"Cleaned output → {self.cleaned_dir}")
+        console.print(f"Removed output → {self.removed_dir}")
 
     # ================================================================
-    # Summary
+    # Summary and Report
     # ================================================================
 
     def _summary(self):
@@ -96,48 +132,79 @@ class CleanFrame:
         kept = len(self.cleaned_paths)
         removed = len(self.removed_paths)
 
-        table = Table(title="CLEANFRAMES SUMMARY", show_header=False, show_lines=False)
-        table.add_row("Model", self.model.capitalize())
-        table.add_row("Clustering", self.cluster_method.upper())
-        table.add_row("Device", str(self.device))
-        table.add_row("Threshold", str(self.threshold))
-        table.add_row("------------------------------------", "")
-        table.add_row("Total frames", str(total))
-        table.add_row("Kept", f"{kept} ({(kept/total)*100:.1f}%)")
-        table.add_row("Removed", f"{removed} ({(removed/total)*100:.1f}%)")
-        table.add_row("------------------------------------", "")
-        table.add_row("Output", self.cleaned_dir)
-        table.add_row("Report", "cf.report() — detailed visual report")
-        console.print(table)
+        lines = [
+            "CleanFrames Summary",
+            "-" * 40,
+            f"Model: {self.model}",
+            f"Clustering: {self.cluster_method}",
+            f"Device: {self.device}",
+            f"Threshold: {self.threshold}",
+            "-" * 40,
+            f"Total frames: {total}",
+            f"Kept: {kept} ({(kept/total)*100:.1f}%)",
+            f"Removed: {removed} ({(removed/total)*100:.1f}%)",
+            "-" * 40,
+            f"Cleaned Output: {self.cleaned_dir}",
+            f"Removed Output: {self.removed_dir}",
+            f"Report: {self.report_path}",
+            "-" * 40
+        ]
+
+        console.print("\n".join(lines))
+        with open(self.report_path, "w") as f:
+            f.write("\n".join(lines))
 
     # ================================================================
     # Main pipeline
     # ================================================================
 
     def run(self):
-        console.rule(f"[bold green]CLEANFRAME PIPELINE — {self.dataset_name}")
-        images, paths = load_images(self.path)
-        self.paths = paths
-        self.embeddings = self.extractor.get_embeddings(images, self.model)
+        console.print(f"\nRunning CleanFrame on '{self.dataset_name}'")
+        self.embeddings, self.paths = self._get_or_load_embeddings()
         self.labels = self._cluster(self.embeddings)
         self._clean(self.embeddings, self.paths, self.labels)
         self._summary()
-        console.rule("[bold green]Done[/bold green]")
 
     # ================================================================
-    # Report (Notebook Visual)
+    # Visualization
     # ================================================================
+    
+    def visualize_clusters(self, method="pca", sample_size=2000):
+        """
+        Visualize embeddings and cluster assignments using PCA or t-SNE.
+        """
+        if self.embeddings is None or self.labels is None:
+            raise ValueError("Run the pipeline first (cf.run()) before visualizing clusters.")
+        
+        # Sample subset if large
+        n = min(sample_size, len(self.embeddings))
+        idx = np.random.choice(len(self.embeddings), n, replace=False)
+        emb = self.embeddings[idx]
+        labels = self.labels[idx]
+
+        # Reduce dimensionality
+        if method == "pca":
+            reducer = PCA(n_components=2)
+            reduced = reducer.fit_transform(emb)
+        elif method == "tsne":
+            reducer = TSNE(n_components=2, random_state=42, perplexity=30)
+            reduced = reducer.fit_transform(emb)
+        else:
+            raise ValueError("method must be 'pca' or 'tsne'.")
+
+        # Plot
+        plt.figure(figsize=(8, 6))
+        plt.scatter(reduced[:, 0], reduced[:, 1], c=labels, cmap='tab10', s=15)
+        plt.title(f"Embedding Visualization ({method.upper()})")
+        plt.xlabel("Component 1")
+        plt.ylabel("Component 2")
+        plt.show()
+
 
     def report(self):
-        console.print("[bold magenta]CleanFrames Detailed Report[/bold magenta]")
-        console.print("---------------------------")
-        console.print(f"Model: {self.model.capitalize()}")
-        console.print(f"Clustering: {self.cluster_method.upper()}")
-        console.print(f"Device: {self.device}")
-        console.print(f"Threshold: {self.threshold}")
-        console.print("------------------------------------")
+        console.print("\nVisual Report (CleanFrames)")
+        console.print("-" * 40)
 
-        # pick 3 random pairs from kept frames
         pairs = []
         if len(self.cleaned_paths) >= 6:
             sample_idx = random.sample(range(len(self.cleaned_paths)), 6)
@@ -157,20 +224,4 @@ class CleanFrame:
             for ax in axes:
                 ax.axis("off")
             fig.suptitle(f"Set {i+1} — Similarity: {sim*100:.1f}%", fontsize=12)
-            plt.show()
-
-            console.print(f"Set {i+1}: Similarity {sim*100:.1f}%")
-            console.print(f"  Image 1: {os.path.basename(p1)}")
-            console.print(f"  Image 2: {os.path.basename(p2)}\n")
-
-        total = len(self.paths)
-        kept = len(self.cleaned_paths)
-        removed = len(self.removed_paths)
-        kept_pct = (kept / total) * 100
-        removed_pct = 100 - kept_pct
-
-        console.print("Summary:")
-        console.print(f"Cleaned with {self.model} embedding and {self.cluster_method}.")
-        console.print(f"{kept_pct:.1f}% kept")
-        console.print(f"{removed_pct:.1f}% removed")
-        console.rule()
\ No newline at end of file
+            plt.show()
\ No newline at end of file
